#!/bin/sh
# Rpm installer object
#
# This file is saved on the disk when the binary .rpm package is installed
# by rpm. It can be invoked by any installer.

#
# Tools
#

# BEGINNING_OF_TMPDIR_DOT_SH
#!/bin/sh

# Create a temporary directory
#
# They are a lot of small utility programs to create temporary files in a
# secure way, but none of them is standard. So I wrote this --hpreg
make_tmp_dir() {
  local dirname="$1" # OUT
  local prefix="$2"  # IN
  local tmp
  local serial
  local loop

  tmp="${TMPDIR:-/tmp}"

  # Don't overwrite existing user data
  # -> Create a directory with a name that didn't exist before
  #
  # This may never succeed (if we are racing with a malicious process), but at
  # least it is secure
  serial=0
  loop='yes'
  while [ "$loop" = 'yes' ]; do
    # Check the validity of the temporary directory. We do this in the loop
    # because it can change over time
    if [ ! -d "$tmp" ]; then
      echo 'Error: "'"$tmp"'" is not a directory.'
      echo
      exit 1
    fi
    if [ ! -w "$tmp" -o ! -x "$tmp" ]; then
      echo 'Error: "'"$tmp"'" should be writable and executable.'
      echo
      exit 1
    fi

    # Be secure
    # -> Don't give write access to other users (so that they can not use this
    # directory to launch a symlink attack)
    if mkdir -m 0755 "$tmp"'/'"$prefix$serial" >/dev/null 2>&1; then
      loop='no'
    else
      serial=`expr "$serial" + 1`
      if [ "`expr "$serial" % 200`" = '0' ]; then
        echo 'Warning: The "'"$tmp"'" directory may be under attack.'
        echo
      fi
    fi
  done

  eval "$dirname"'="$tmp"'"'"'/'"'"'"$prefix$serial"'
}

# END_OF_TMPDIR_DOT_SH

# BEGINNING_OF_DB_DOT_SH
#!/bin/sh

#
# Manage an installer database
#

# Add an answer to a database in memory
db_answer_add() {
  local dbvar="$1" # IN/OUT
  local id="$2"    # IN
  local value="$3" # IN
  local answers
  local i

  eval "$dbvar"'_answer_'"$id"'="$value"'

  eval 'answers="$'"$dbvar"'_answers"'
  # There is no double quote around $answers on purpose
  for i in $answers; do
    if [ "$i" = "$id" ]; then
      return
    fi
  done
  answers="$answers"' '"$id"
  eval "$dbvar"'_answers="$answers"'
}

# Remove an answer from a database in memory
db_answer_remove() {
  local dbvar="$1" # IN/OUT
  local id="$2"    # IN
  local new_answers
  local answers
  local i

  eval 'unset '"$dbvar"'_answer_'"$id"

  new_answers=''
  eval 'answers="$'"$dbvar"'_answers"'
  # There is no double quote around $answers on purpose
  for i in $answers; do
    if [ "$i" != "$id" ]; then
      new_answers="$new_answers"' '"$i"
    fi
  done
  eval "$dbvar"'_answers="$new_answers"'
}

# Load all answers from a database on stdin to memory (<dbvar>_answer_*
# variables)
db_load_from_stdin() {
  local dbvar="$1" # OUT

  eval "$dbvar"'_answers=""'
  while read -r action p1 p2; do
    if [ "$action" = 'answer' ]; then
      db_answer_add "$dbvar" "$p1" "$p2"
    elif [ "$action" = 'remove_answer' ]; then
      db_answer_remove "$dbvar" "$p1"
    fi
  done
}

# Load all answers from a database on disk to memory (<dbvar>_answer_*
# variables)
db_load() {
  local dbvar="$1"  # OUT
  local dbfile="$2" # IN

  db_load_from_stdin "$dbvar" < "$dbfile"
}

# Iterate through all answers in a database in memory, calling <func> with
# id/value pairs and the remaining arguments to this function
db_iterate() {
  local dbvar="$1" # IN
  local func="$2"  # IN
  shift 2
  local answers
  local i
  local value

  eval 'answers="$'"$dbvar"'_answers"'
  # There is no double quote around $answers on purpose
  for i in $answers; do
    eval 'value="$'"$dbvar"'_answer_'"$i"'"'
    "$func" "$i" "$value" "$@"
  done
}

# If it exists in memory, remove an answer from a database (disk and memory)
db_remove_answer() {
  local dbvar="$1"  # IN/OUT
  local dbfile="$2" # IN
  local id="$3"     # IN
  local answers
  local i

  eval 'answers="$'"$dbvar"'_answers"'
  # There is no double quote around $answers on purpose
  for i in $answers; do
    if [ "$i" = "$id" ]; then
      echo 'remove_answer '"$id" >> "$dbfile"
      db_answer_remove "$dbvar" "$id"
      return
    fi
  done
}

# Add an answer to a database (disk and memory)
db_add_answer() {
  local dbvar="$1"  # IN/OUT
  local dbfile="$2" # IN
  local id="$3"     # IN
  local value="$4"  # IN

  db_remove_answer "$dbvar" "$dbfile" "$id"
  echo 'answer '"$id"' '"$value" >> "$dbfile"
  db_answer_add "$dbvar" "$id" "$value"
}

# Add a file to a database on disk
# 'file' is the file to put in the database (it may not exist on the disk)
# 'tsfile' is the file to get the timestamp from, '' if no timestamp
db_add_file() {
  local dbfile="$1" # IN
  local file="$2"   # IN
  local tsfile="$3" # IN
  local date

  if [ "$tsfile" = '' ]; then
    echo 'file '"$file" >> "$dbfile"
  else
    date=`date -r "$tsfile" '+%s' 2> /dev/null`
    if [ "$date" != '' ]; then
      date=' '"$date"
    fi
    echo 'file '"$file$date" >> "$dbfile"
  fi
}

# Add a directory to a database on disk
db_add_dir() {
  local dbfile="$1" # IN
  local dir="$2"    # IN

  echo 'directory '"$dir" >> "$dbfile"
}
# END_OF_DB_DOT_SH

#
# Implementation of the methods
#

# Return the human-readable type of the installer
installer_kind() {
  echo 'rpm'

  exit 0
}

# Return the human-readable version of the installer
installer_version() {
  echo '3'

  exit 0
}

# Return the specific VMware product
vmware_product() {
  echo 'ws'

  exit 0
}

# this is a function and not just a macro because someday the product
# name may contain language-specific escape characters.  (talk to hpreg)
vmware_product_name() {
  echo 'VMware Workstation'
  exit 0
}
		
# Set the name of the main /etc/vmware* directory
# Set up variables depending on the main RegistryDir
initialize_globals() {
  if [ "`vmware_product`" = 'console' ]; then
    gRegistryDir='/etc/vmware-console'
    gProductName='VMware-console'
  elif [ "`vmware_product`" = 'wgs' ]; then
    gRegistryDir='/etc/vmware'
    gProductName='VMware-gsx'
  else
    gRegistryDir='/etc/vmware'
    gProductName='VMwareWorkstation'
  fi
  gInstallerMainDB="$gRegistryDir"'/locations'
}

# Convert the installer database format to formats used by older installers
# The output should be a .tar.gz containing enough information to allow a
# clean "upgrade" (which will actually be a downgrade) by an older installer
installer_convertdb() {
  local format="$1"
  local output="$2"
  local tmpdir

  case "$format" in
    rpm3|tar3)
      if [ "$format" = 'rpm3' ]; then
        echo 'Keeping the rpm3 installer database format.'
      else
        echo 'Converting the rpm3 installer database format'
        echo '        to the tar3 installer database format.'
      fi
      echo
      # The next installer uses the same database format. Backup a full
      # database state that it can use as a fresh new database.
      #
      # Everything should go in:
      #  /etc/vmware*/
      #              state/
      #
      # But those directories do not have to be referenced,
      # the next installer will do it just after restoring the backup
      # because only it knows if those directories have been created.
      #
      # Also, do not include those directories in the backup, because some
      # versions of tar (1.13.17+ are ok) do not untar directory permissions
      # as described in their documentation.
      make_tmp_dir 'tmpdir' 'vmware-installer'
      mkdir -p "$tmpdir""$gRegistryDir"
      db_add_file "$tmpdir""$gInstallerMainDB" "$gInstallerMainDB" ''
      db_load 'db' "$gInstallerMainDB"
      write() {
        local id="$1"
        local value="$2"
        local dbfile="$3"

        echo 'answer '"$id"' '"$value" >> "$dbfile"
      }
      db_iterate 'db' 'write' "$tmpdir""$gInstallerMainDB"
      files='./'"$gInstallerMainDB"
      if [ "$db_answer_VNET_1_SAMBA" = 'yes' ]; then
        mkdir -p "$tmpdir""$gRegistryDir"'/state'
        cp "$gRegistryDir"/vmnet1/smb/private/smbpasswd "$tmpdir""$gRegistryDir"'/state/1'
        db_add_file "$tmpdir""$gInstallerMainDB" "$gRegistryDir"'/state/1' "$tmpdir""$gRegistryDir"'/state/1'
        echo 'answer VNET_1_SAMBA_SMBPASSWD '"$gRegistryDir"'/state/1' >> "$tmpdir""$gInstallerMainDB"
        files="$files"' .'"$gRegistryDir"'/state/1'
        cp "$gRegistryDir"/vmnet1/smb/private/MACHINE.SID "$tmpdir""$gRegistryDir"'/state/2'
        db_add_file "$tmpdir""$gInstallerMainDB" "$gRegistryDir"'/state/2' "$tmpdir""$gRegistryDir"'/state/2'
        echo 'answer VNET_1_SAMBA_MACHINESID '"$gRegistryDir"'/state/2' >> "$tmpdir""$gInstallerMainDB"
        files="$files"' .'"$gRegistryDir"'/state/2'
      fi
      if [ "`vmware_product`" = 'server' ]; then
        # Save file used by management UI
        if [ -e "$gRegistryDir"/vm-list ]; then
          mkdir -p "$tmpdir""$gRegistryDir"'/state'
          cp "$gRegistryDir"/vm-list "$tmpdir""$gRegistryDir"'/state/3'
         db_add_file "$tmpdir""$gInstallerMainDB" "$gRegistryDir"'/state/3' "$tmpdir""$gRegistryDir"'/state/3'
         echo 'answer VSS_VM_LIST '"$gRegistryDir"'/state/3' >> "$tmpdir""$gInstallerMainDB"
         files="$files"' .'"$gRegistryDir"'/state/3'
        fi

        # Save files used by scalable server installer
        if [ -e "$gRegistryDir"/hwconfig ]; then
          mkdir -p "$tmpdir""$gRegistryDir"'/state'
         cp "$gRegistryDir"/hwconfig "$tmpdir""$gRegistryDir"'/state/4'
         db_add_file "$tmpdir""$gInstallerMainDB" "$gRegistryDir"'/state/4' "$tmpdir""$gRegistryDir"'/state/4'
         echo 'answer VSS_HWCONF_SAV '"$gRegistryDir"'/state/4' >> "$tmpdir""$gInstallerMainDB"
         files="$files"' .'"$gRegistryDir"'/state/4'
        fi
      fi
      # There is no double quote around $files on purpose
      tar -C "$tmpdir" -czopf "$output" $files 2> /dev/null
      rm -rf "$tmpdir"

      exit 0
      ;;

    rpm2|tar2)
      echo 'Converting the rpm3 installer database format'
      echo '        to the '"$format"' installer database format.'
      echo
      # The next installer uses the same database format. Backup a full
      # database state that it can use as a fresh new database.
      #
      # Everything should go in:
      #  /etc/vmware*/
      #              state/
      #
      # But those directories do not have to be referenced,
      # the next installer will do it just after restoring the backup
      # because only it knows if those directories have been created.
      #
      # Also, do not include those directories in the backup, because some
      # versions of tar (1.13.17+ are ok) do not untar directory permissions
      # as described in their documentation.
      make_tmp_dir 'tmpdir' 'vmware-installer'
      mkdir -p "$tmpdir""$gRegistryDir"
      db_add_file "$tmpdir""$gInstallerMainDB" "$gInstallerMainDB" ''
      db_load 'db' "$gInstallerMainDB"
      write() {
        local id="$1"
        local value="$2"
        local dbfile="$3"

        # For the rpm3|tar3 format, a number of keywords were removed.  In their 
        # place a more flexible scheme was implemented for which each has a semantic
        # equivalent:
	#
        #   VNET_SAMBA             -> VNET_1_SAMBA
        #   VNET_SAMBA_MACHINESID  -> VNET_1_SAMBA_MACHINESID
        #   VNET_SAMBA_SMBPASSWD   -> VNET_1_SAMBA_SMBPASSWD
        #   VNET_HOSTONLY          -> VNET_1_HOSTONLY
        #   VNET_HOSTONLY_HOSTADDR -> VNET_1_HOSTONLY_HOSTADDR
        #   VNET_HOSTONLY_NETMASK  -> VNET_1_HOSTONLY_NETMASK
        #   VNET_INTERFACE         -> VNET_0_INTERFACE
	# 
	# We undo the changes from rpm2|tar2 to rpm3|tar3.
        if [ "$id" = 'VNET_1_SAMBA' ]; then
          id='VNET_SAMBA'
        elif [ "$id" = 'VNET_1_SAMBA_MACHINESID' ]; then
          id='VNET_SAMBA_MACHINESID'
        elif [ "$id" = 'VNET_1_SAMBA_SMBPASSWD' ]; then
          id='VNET_SAMBA_SMBPASSWD'
        elif [ "$id" = 'VNET_1_HOSTONLY' ]; then
          id='VNET_HOSTONLY'
        elif [ "$id" = 'VNET_1_HOSTONLY_HOSTADDR' ]; then
          id='VNET_HOSTONLY_HOSTADDR'
        elif [ "$id" = 'VNET_1_HOSTONLY_NETMASK' ]; then
          id='VNET_HOSTONLY_NETMASK'
        elif [ "$id" = 'VNET_0_INTERFACE' ]; then
          id='VNET_INTERFACE'
        fi

        echo 'answer '"$id"' '"$value" >> "$dbfile"
      }
      db_iterate 'db' 'write' "$tmpdir""$gInstallerMainDB"
      files='.'"$gInstallerMainDB"
      # We ignore all the other possible Samba servers because rpm2|tar2 
      # only support 1 samba server
      # XXX Should we save the state for the other Samba servers?
      if [ "$db_answer_VNET_1_SAMBA" = 'yes' ]; then
        mkdir -p "$tmpdir""$gRegistryDir"'/state'
        cp "$gRegistryDir"/vmnet1/smb/private/smbpasswd "$tmpdir""$gRegistryDir"'/state/1'
        db_add_file "$tmpdir""$gInstallerMainDB" "$gRegistryDir"'/state/1' "$tmpdir""$gRegistryDir"'/state/1'
        echo 'answer VNET_SAMBA_SMBPASSWD '"$gRegistryDir"'/state/1' >> "$tmpdir""$gInstallerMainDB"
        files="$files"' .'"$gRegistryDir"'/state/1'
        cp "$gRegistryDir"/vmnet1/smb/private/MACHINE.SID "$tmpdir""$gRegistryDir"'/state/2'
        db_add_file "$tmpdir""$gInstallerMainDB" "$gRegistryDir"'/state/2' "$tmpdir""$gRegistryDir"'/state/2'
        echo 'answer VNET_SAMBA_MACHINESID '"$gRegistryDir"'/state/2' >> "$tmpdir""$gInstallerMainDB"
        files="$files"' .'"$gRegistryDir"'/state/2'
      fi
      # There is no double quote around $files on purpose
      tar -C "$tmpdir" -czopf "$output" $files 2> /dev/null
      rm -rf "$tmpdir"

      exit 0
      ;;

    rpm|tar)
      echo 'Converting the rpm3 installer database format'
      echo '        to the '"$format"'  installer database format.'
      echo
      # Backup only the main database file. The next installer ignores
      # new keywords as well as file and directory statements, and deals
      # properly with remove_ statements
      tar -C '/' -czopf "$output" '.'"$gInstallerMainDB" 2> /dev/null

      exit 0
      ;;

    old)
      echo 'Converting the rpm3 installer database format'
      echo '        to the old  installer database format.'
      echo
      # Backup only the main database file (there is no need to reference
      # it). Remove all file and directory statements because the
      # uninstallation will be handled by the uninstall method in this
      # file. The next installer ignore other keywords. Unfortunately,
      # answer statements are read (sometimes wrongly, because the next
      # installer doesn't know about remove_ statements) before this is
      # executed. Basically, we can only help to make the old uninstaller
      # look good here.
      make_tmp_dir 'tmpdir' 'vmware-installer'
      mkdir -p "$tmpdir""$gRegistryDir"
      echo 'VERSION 3.1.1' >> "$tmpdir""$gInstallerMainDB"
      tar -C "$tmpdir" -czopf "$output" '.'"$gInstallerMainDB" 2> /dev/null
      rm -rf "$tmpdir"

      exit 0
      ;;

    *)
      echo 'Unknown '"$format"' installer database format.'
      echo

      exit 1
      ;;
  esac
}

# This friend method should be called directly only by installers that know
# exactly what they do (typically only the rpm installer).
installer_uninstall_without_rpm() {
  local display="$1"

  # Unconfigure the package
  perl -w - << 'END_OF_UNCONFIG_DOT_PL'
#!/usr/bin/perl -w
# If your copy of perl is not in /usr/bin, please adjust the line above.
#
# Copyright (C) 1998-2000 VMware, Inc.  All Rights Reserved.
#
# Unconfigurator for VMware

use strict;

# External helper programs
my %gHelper;

# Global variables
my %gDBAnswer;
my %gDBFile;
my %gDBDir;
my $gInstallerMainDB;

# Set the name of the main /etc/vmware* directory.
sub initialize_globals {
  if (vmware_product() eq 'console') {
    $gInstallerMainDB = '/etc/vmware-console/locations';
  } else {
    $gInstallerMainDB = '/etc/vmware/locations';
  }
}

# Load the installer database
sub db_load {
  undef %gDBAnswer;
  undef %gDBFile;
  undef %gDBDir;

  open(INSTALLDB, '<' . $gInstallerMainDB) or error('Unable to open the installer database ' . $gInstallerMainDB . ' in read-mode.' . "\n\n");
  while (<INSTALLDB>) {
    if (/^answer (\S+) (\S+)/) {
      $gDBAnswer{$1} = $2;
    } elsif (/^answer (\S+)/) {
      $gDBAnswer{$1} = '';
    } elsif (/^remove_answer (\S+)/) {
      delete $gDBAnswer{$1};
    } elsif (/^file (\S+) (\S+)/) {
      $gDBFile{$1} = $2;
    } elsif (/^file (\S+)/) {
      $gDBFile{$1} = 0;
    } elsif (/^remove_file (\S+)/) {
      delete $gDBFile{$1};
    } elsif (/^directory (\S+)/) {
      $gDBDir{$1} = '';
    } elsif (/^remove_directory (\S+)/) {
      delete $gDBDir{$1};
    }
  }
  close(INSTALLDB);
}

# Open the database on disk in append mode
sub db_append {
  if (not open(INSTALLDB, '>>' . $gInstallerMainDB)) {
    error('Unable to open the installer database ' . $gInstallerMainDB . ' in append-mode.' . "\n\n");
  }
  # Force a flush after every write operation.
  # See 'Programming Perl', p. 110
  select((select(INSTALLDB), $| = 1)[0]);
}

# Remove a file from the tar installer database
sub db_remove_file {
  my $file = shift;

  print INSTALLDB 'remove_file ' . $file . "\n";
  delete $gDBFile{$file};
}

# Remove a directory from the tar installer database
sub db_remove_dir {
  my $dir = shift;

  print INSTALLDB 'remove_directory ' . $dir . "\n";
  delete $gDBDir{$dir};
}

# Determine if a file belongs to the tar installer database
sub db_file_in {
  my $file = shift;

  return defined($gDBFile{$file});
}

# Determine if a directory belongs to the tar installer database
sub db_dir_in {
  my $dir = shift;

  return defined($gDBDir{$dir});
}

# Return the timestamp of an installed file
sub db_file_ts {
  my $file = shift;

  return $gDBFile{$file};
}

# Retrieve an answer that must be present in the database
sub db_get_answer {
  my $id = shift;

  if (not defined($gDBAnswer{$id})) {
    error('Unable to find the answer ' . $id . ' in the installer database (' . $gInstallerMainDB . '). You may want to re-install ' . vmware_product_name() . '.\n\n');
  }

  return $gDBAnswer{$id};
}

# Save the tar installer database
sub db_save {
  close(INSTALLDB);
}

# Constants
my $cTerminalLineSize = 80;

# Wordwrap system: append some content to the output
sub append_output {
  my $output = shift;
  my $pos = shift;
  my $append = shift;

  $output .= $append;
  $pos += length($append);
  if ($pos >= $cTerminalLineSize) {
    $output .= "\n";
    $pos = 0;
  }

  return ($output, $pos);
}

# Wordwrap system: deal with the next character
sub wrap_one_char {
  my $output = shift;
  my $pos = shift;
  my $word = shift;
  my $char = shift;
  my $reserved = shift;
  my $length;

  if (not (($char eq "\n") || ($char eq ' ') || ($char eq ''))) {
    $word .= $char;

    return ($output, $pos, $word);
  }

  # We found a separator. Process the last word

  $length = length($word) + $reserved;
  if (($pos + $length) > $cTerminalLineSize) {
    # The last word doesn't fit in the end of the line. Break the line before it
    $output .= "\n";
    $pos = 0;
  }
  ($output, $pos) = append_output($output, $pos, $word);
  $word = '';

  if ($char eq "\n") {
    $output .= "\n";
    $pos = 0;
  } elsif ($char eq ' ') {
    if ($pos) {
      ($output, $pos) = append_output($output, $pos, ' ');
    }
  }

  return ($output, $pos, $word);
}

# Wordwrap system: word-wrap a string plus some reserved trailing space
sub wrap {
  my $input = shift;
  my $reserved = shift;
  my $output;
  my $pos;
  my $word;
  my $i;

  $output = '';
  $pos = 0;
  $word = '';
  for ($i = 0; $i < length($input); $i++) {
    ($output, $pos, $word) = wrap_one_char($output, $pos, $word, substr($input, $i, 1), 0);
  }
  # Use an artifical last '' separator to process the last word
  ($output, $pos, $word) = wrap_one_char($output, $pos, $word, '', $reserved);

  return $output;
}

# Print an error message and exit
sub error {
  my $msg = shift;

  print STDERR wrap($msg . 'Execution aborted.' . "\n\n", 0);
  exit 1;
}

# Convert a string to its equivalent shell representation
sub shell_string {
  my $single_quoted = shift;

  $single_quoted =~ s/'/'"'"'/g;
  # This comment is a fix for emacs's broken syntax-highlighting code --hpreg
  return '\'' . $single_quoted . '\'';
}

# Contrary to a popular belief, 'which' is not always a shell builtin command.
# So we can not trust it to determine the location of other binaries.
# Moreover, SuSE 6.1's 'which' is unable to handle program names beginning with
# a '/'...
#
# Return value is the complete path if found, or '' if not found
sub internal_which {
  my $bin = shift;

  if (substr($bin, 0, 1) eq '/') {
    # Absolute name
    if ((-f $bin) && (-x $bin)) {
      return $bin;
    }
  } else {
    # Relative name
    my @paths;
    my $path;

    if (index($bin, '/') == -1) {
      # There is no other '/' in the name
      @paths = split(':', $ENV{'PATH'});
      foreach $path (@paths) {
   my $fullbin;

   $fullbin = $path . '/' . $bin;
   if ((-f $fullbin) && (-x $fullbin)) {
     return $fullbin;
   }
      }
    }
  }

  return '';
}

# Emulate a simplified sed program
# Return 1 if success, 0 if failure
# XXX as a side effect, if the string being replaced is '', remove
# the entire line.  Remove this, once we have better "block handling" of
# our config data in config files.
sub internal_sed {
  my $src = shift;
  my $dst = shift;
  my $append = shift;
  my $patchRef = shift;
  my @patchKeys;

  if (not open(SRC, '<' . $src)) {
    return 0;
  }
  if (not open(DST, (($append == 1) ? '>>' : '>') . $dst)) {
    return 0;
  }

  @patchKeys = keys(%$patchRef);
  if ($#patchKeys == -1) {
    while(defined($_ = <SRC>)) {
      print DST $_;
    }
  } else {
    while(defined($_ = <SRC>)) {
      my $patchKey;
      my $del = 0;

      foreach $patchKey (@patchKeys) {
      if (s/$patchKey/$$patchRef{$patchKey}/g) {
          if ($_ eq "\n") {
            $del = 1;
          }
        }
      }
      next if ($del);
      print DST $_;
    }
  }

  close(SRC);
  close(DST);
  return 1;
}

# Check if a file name exists
sub file_name_exist {
  my $file = shift;

  # Note: We must test for -l before, because if an existing symlink points to
  #       a non-existing file, -e will be false
  return ((-l $file) || (-e $file))
}

# Find a suitable backup name and backup a file
sub backup_file {
  my $file = shift;
  my $i;

  for ($i = 0; $i < 100; $i++) {
    if (not file_name_exist($file . '.old.' . $i)) {
      my %patch;

      undef %patch;
      if (internal_sed($file, $file . '.old.' . $i, 0, \%patch)) {
         print wrap('File ' . $file . ' is backed up to ' . $file . '.old.' .
         $i . '.' . "\n\n", 0);
      } else {
         print STDERR wrap('Unable to backup the file ' . $file . ' to ' .
         $file . '.old.' . $i .'.' . "\n\n", 0);
      }
      return;
    }
  }

  print STDERR wrap('Unable to backup the file ' . $file .
      '. You have too many backups files. They are files of the form ' .
      $file . '.old.N, where N is a number. Please delete some of them.' .
      "\n\n", 0);
}

# Uninstall a file previously installed by us
sub uninstall_file {
  my $file = shift;

  if (not db_file_in($file)) {
    # Not installed by this script
    return;
  }

  if (file_name_exist($file)) {
    if (db_file_ts($file)) {
      my @statbuf;

      @statbuf = stat($file);
      if (defined($statbuf[9])) {
   if (db_file_ts($file) != $statbuf[9]) {
     # Modified since this script installed it
     backup_file($file);
   }
      } else {
         print STDERR wrap('Unable to get the last modification timestamp of the file ' .
            $file . '.' . "\n\n", 0);
      }
    }

    if (not unlink($file)) {
      print STDERR wrap('Unable to remove the file ' . $file . '.' . "\n\n", 0);
    }
  } else {
      print wrap('This script previously created the file ' . $file .
      ', and was about to remove it. Somebody else apparently did it already.' . "\n\n", 0);
  }

  db_remove_file($file);
}

# Uninstall a directory previously installed by us
sub uninstall_dir {
  my $dir = shift;

  if (not db_dir_in($dir)) {
    # Not installed by this script
    return;
  }

  if (-d $dir) {
    if (not rmdir($dir)) {
      print wrap('This script previously created the directory ' . $dir . ', and was about to remove it. Since there are files in that directory that this script did not create, it will not be removed.' . "\n\n", 0);
      if (   defined($ENV{'VMWARE_DEBUG'})
          && ($ENV{'VMWARE_DEBUG'} eq 'yes')) {
        system('ls -AlR ' . shell_string($dir));
      }
    }
  } else {
    print wrap('This script previously created the directory ' . $dir . ', and was about to remove it. Somebody else apparently did it already.' . "\n\n", 0);
  }

  db_remove_dir($dir);
}

# Uninstall files and directories beginning with a given prefix
sub uninstall_prefix {
  my $prefix = shift;
  my $prefix_len;
  my $file;
  my $dir;

  $prefix_len = length($prefix);

  # Remove all files beginning with $prefix
  foreach $file (keys %gDBFile) {
    if (substr($file, 0, $prefix_len) eq $prefix) {
      uninstall_file($file);
    }
  }

  # Remove all directories beginning with $prefix
  # We sort them by decreasing order of their length, to ensure that we will
  # remove the inner ones before the outer ones
  foreach $dir (sort {length($b) <=> length($a)} keys %gDBDir) {
    if (substr($dir, 0, $prefix_len) eq $prefix) {
      uninstall_dir($dir);
    }
  }
}

# Unconfigure a rpm package
sub uninstall {
  if (defined($gDBAnswer{'INITSCRIPTSDIR'}) && db_file_in(db_get_answer('INITSCRIPTSDIR') . '/vmware')) {
    # The installation process ran far enough to create the startup script
    my $status;

    # In case service links were created the LSB way, remove them
    if (not ($gHelper{'insserv'} eq '')) {
      system(shell_string($gHelper{'insserv'}) . ' -r ' . shell_string(db_get_answer('INITSCRIPTSDIR') . '/vmware'));
    }

    # Stop the services
    $status = system(shell_string(db_get_answer('INITSCRIPTSDIR') . '/vmware') . ' stop') >> 8;
    if ($status) {
      if ($status == 2) {
        # At least one instance of VMware is still running. We must refuse to
        # uninstall
        error('Unable to stop ' . vmware_product_name() . '\'s services. Aborting the uninstallation.' . "\n\n");
      }

      # Oh well, at worst the user will have to reboot the machine... The
      # uninstallation process should go as far as possible
      print STDERR wrap('Unable to stop ' . vmware_product_name() . '\'s services.' . "\n\n", 0);
    } else {
      print "\n";
    }
  }

  uninstall_prefix('');
}

# Return the specific VMware product
sub vmware_product {
  return 'ws';
}

# this is a function and not a macro in case the product name ever contains
# language-specific escape characters (ask hpreg)
sub vmware_product_name {
  return 'VMware Workstation'
}

# Set up the location of external helpers
sub initialize_external_helpers {
  $gHelper{'insserv'} = internal_which('insserv');
}

# If this was a WGS build, remove our inetd.conf entry for auth daemon
# and stop the vmware-serverd
sub wgs_uninstall {
  system('killall -TERM vmware-serverd  >/dev/null 2>&1'); 
  uninstall_superserver();
}

# Try and figure out which "superserver" is installed and unconfigure correct
# one.
sub uninstall_superserver {
  my $inetd_conf  = "/etc/inetd.conf";
  my $xinetd_dir  = "/etc/xinetd.d";

  # check for xinetd
  # XXX Could be a problem, as they could start xinetd with '-f config_file'.
  #     We could do a ps -ax, look for xinetd, parse the line, find the config
  #     file, parse the config file to find the xinet.d directory.  Or parse if
  #     from the init.d script somewhere.  If they use init.d.
  if (-d $xinetd_dir) {
    uninstall_xinetd($xinetd_dir);
    return;
  }

  # check for inetd
  if (not (vmware_product() eq 'server')) {
    if (-e $inetd_conf) {
      uninstall_inetd($inetd_conf);
      return;
    }
  }
}

# Cleanup the inetd.conf file.
sub uninstall_inetd {
  my $inetd = "/etc/inetd.conf";
  my %patch = ('^# VMware auth.*$' => '',
          '^.*stream\s+tcp\s+nowait.*vmauthd.*$' => '',
          '^.*stream\s+tcp\s+nowait.*vmware-authd.*$' => '');
  my $tmp_dir = make_tmp_dir('vmware-installer');
  # Build the temp file's path
  my $tmp = $tmp_dir . '/inetd.' . $$;

  internal_sed($inetd, $tmp, 0, \%patch);
  undef %patch;

  if (not internal_sed($tmp, $inetd, 0, \%patch)) {
    print STDERR wrap('Unable to copy file ' . $tmp . ' back to ' . $inetd . '.' . "\n" . 'The authentication daemon was not removed from ' . $inetd . "\n\n", 0);
  }
  unlink($tmp);
  unlink($tmp_dir);
  system('killall -HUP inetd'); 
}

# Cleanup the xinetd.d directory.
sub uninstall_xinetd {
  my $conf_dir  = shift;

  unlink($conf_dir . '/vmware-authd');
  system('killall -USR2 xinetd'); 

}

# BEGINNING_OF_TMPDIR_DOT_PL
# END_OF_TMPDIR_DOT_PL

# Program entry point
sub main {
    # It is actually 'unconfiguring', but we don't want to confuse the user
      print wrap('Uninstalling the rpm installation of ' . vmware_product_name() . '.\n\n', 0);

    # Force the path to reduce the risk of using "modified" external helpers
    # If the user has a special system setup, he will will prompted for the
    # proper location anyway
    $ENV{'PATH'} = '/bin:/usr/bin:/sbin:/usr/sbin';

    initialize_globals();
    initialize_external_helpers();

    db_load();
    db_append();

    if ((vmware_product() eq 'wgs') || (vmware_product() eq 'server')) {
      wgs_uninstall();
    }

    uninstall();
    db_save();

    exit 0;
}

main();
END_OF_UNCONFIG_DOT_PL
  if [ ! "$?" -eq 0 ]; then
    exit 1
  fi

  if [ "$display" = "yes" ]; then
    echo 'The removal of '"`vmware_product_name`"' 3.1.1 for Linux completed successfully.'
    echo 'Thank you for having tried this software.'
    echo
  fi

  exit 0
}

# Uninstall what has been installed by the installer
# This should never prompt the user, because it can be annoying if invoked
# from the rpm installer for example.
installer_uninstall() {
  # Remove the package
  rpm -e $gProductName || exit 1

  exit 0
}

#
# Interface of the methods
#

initialize_globals

case "$1" in
  kind)
    installer_kind
    ;;

  version)
    installer_version
    ;;

  convertdb)
    installer_convertdb "$2" "$3"
    ;;

  uninstall)
    installer_uninstall
    ;;

  uninstall_without_rpm)
    installer_uninstall_without_rpm "$2"
    ;;

  *)
    echo 'Usage: '"`basename "$0"`"' {kind|version|convertdb|uninstall|uninstall_without_rpm}'
    echo

    exit 1
    ;;
esac

